/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is New Tab Page code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

"use strict";

let Cu = Components.utils;
let Ci = Components.interfaces;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/Geometry.jsm");
Cu.import("resource:///modules/NewTabUtils.jsm");

const PREF_NEWTAB_ENABLED = "browser.newtab.enabled";

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is New Tab Page code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO
let Animations = {
  fadeIn: function Anim_fadeIn(aNode, aCallback) {
    new CssAnimation(aNode, "anim-fadein", aCallback);
  },

  fadeOut: function Anim_fadeOut(aNode, aCallback) {
    new CssAnimation(aNode, "anim-fadeout", aCallback);
  }
};

// TODO
function CssAnimation(aNode, aName, aCallback) {
  this._node = aNode;
  this._name = aName;
  this._callback = aCallback;

  this._start();
}

CssAnimation.prototype = {
  _start: function CssAnim__start() {
    let node = this._node;
    let name = this._name;
    let classes = node.classList;
    let callback = this._callback;

    this._whenAnimationEnded(node, function () {
      classes.remove(name);

      if (callback)
        callback();
    });

    classes.add(name);
  },

  _whenAnimationEnded: function CssAnim__whenAnimationEnded(aNode, aCallback) {
    aNode.addEventListener("animationend", function onEnd() {
      aNode.removeEventListener("animationend", onEnd, false);
      aCallback();
    }, false);
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is New Tab Page code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO
let Transformation = {
  blockSite: function Trans_blockSite(aSite, aCallback) {
    return new BlockSiteTransformation(aSite, aCallback);
  },

  unpinSite: function Trans_unpinSite(aSite, aCallback) {
    return new UnpinSiteTransformation(aSite, aCallback);
  },

  dragSite: function Trans_dragSite(aSite, aCell, aCallback) {
    return new DragSiteTransformation(aSite, aCell, aCallback);
  }
};

// TODO
let BaseTransformation = {
  _getNodePosition: function BaseTrans__getNodePosition(aNode) {
    let rect = aNode.getBoundingClientRect();
    return new Rect(scrollX + rect.left, scrollY + rect.top,
                    rect.width, rect.height);
  },

  _setSitePosition: function BaseTrans__setSitePosition(aSite, aPosition) {
    let style = aSite.node.style;
    let {top, left} = aPosition;

    style.top = top + "px";
    style.left = left + "px";
  },

  _freezeSitePosition: function BaseTrans__freezeSitePosition(aSite) {
    this._setSitePosition(aSite, this._getNodePosition(aSite.node));
    aSite.node.style.position = "absolute";
  },

  _unfreezeSitePosition: function BaseTrans__unfreezeSitePosition(aSite) {
    let style = aSite.node.style;
    style.top = style.left = style.position = "";
  },

  _slideSiteTo: function BaseTrans__slideSiteTo(aSite, aTarget, aCallback) {
    this._setSitePosition(aSite, this._getNodePosition(aTarget.node));

    if (aCallback)
      this._whenTransitionEnded(aSite.node, aCallback);
  },

  _whenTransitionEnded: function BaseTrans__whenTransitionEnded(aNode, aCallback) {
    aNode.addEventListener("transitionend", function onEnd() {
      aNode.removeEventListener("transitionend", onEnd, false);
      aCallback();
    }, false);
  },

  _lock: function BaseTrans__lock() {
    NewTabUtils.Pages.lock();
  },

  _unlock: function BaseTrans__unlock() {
    NewTabUtils.Pages.unlock();
  }
};

// TODO
function BlockSiteTransformation(aSite, aCallback) {
  this._site = aSite;
  this._callback = aCallback;

  this._start();
}

BlockSiteTransformation.prototype = extend(BaseTransformation, {
  _start: function BSTrans__start() {
    this._lock();

    let cells = Grid.cells.slice(this._site.index);

    // get all unpinned and non-empty cells
    this._movableCells = cells.filter(function (cell) {
      return !cell.isEmpty() && !cell.containsPinnedSite();
    });

    this._movableSites = this._movableCells.map(function (cell) cell.site);

    // freeze the site positions
    this._movableSites.forEach(this._freezeSitePosition, this);

    // fade out the blocked site and continue with step 2
    Animations.fadeOut(this._site.node, this._step2.bind(this));
  },

  // TODO better name
  _step2: function BSTrans__step2() {
    let movableCells = this._movableCells;
    let movableSites = this._movableSites;

    let numCells = movableCells.length;
    let numSlides = numCells - 1;

    let self = this;

    function slideFinished() {
      if (!--numSlides)
        self._step3();
    }

    for (let i = 1; i < numCells; i++) {
      let targetSite = movableSites[i - 1];
      let targetCell = movableCells[i - 1];
      let site = movableSites[i];

      // move site the its new DOM position
      targetCell.node.appendChild(site.node);

      // slide site into its new position
      this._slideSiteTo(site, targetSite, slideFinished);
    }

    // remove the blocked site's DOM node
    let node = this._site.node;
    node.parentNode.removeChild(node);

    // no cells will be slided around, go to step 3
    if (numCells < 2)
      this._step3();
  },

  // TODO better name
  _step3: function BSTrans__step3() {
    let movableCells = this._movableCells;
    let lastCell = movableCells[movableCells.length - 1];
    let index = Grid.cells.indexOf(lastCell);

    // unfreeze the site positions
    this._movableSites.forEach(this._unfreezeSitePosition, this);

    let self = this;

    // find the next link that will appear
    Links.getLinks(function (links) {
      if (links[index]) {
        let site = Grid.createSite(links[index], lastCell);
        Animations.fadeIn(site.node, self._finish.bind(self));
      } else {
        self._finish();
      }
    });
  },

  _finish: function BSTrans__finish() {
    this._unlock();

    if (this._callback)
      this._callback.callback();
  }
});

// TODO
function UnpinSiteTransformation(aSite, aCallback) {
  this._site = aSite;
  this._callback = aCallback;

  this._start();
}

UnpinSiteTransformation.prototype = extend(BaseTransformation, {
  _start: function USTrans__start() {
    let firstEmptyCell = this._findFirstEmptyCell();

    if (firstEmptyCell) {
      this._moveSiteToEmptyCell(firstEmptyCell);
    } else {
      // no empty cells before the unpinned cell
      this._finish();
    }
  },

  _findFirstEmptyCell: function USTrans__findFirstEmptyCell() {
    let cells = Grid.cells;
    let maxIndex = this._site.index;

    for (let i = 0; i < maxIndex; i++) {
      let cell = cells[i];

      if (cell.isEmpty())
        return cell;
    }

    return null;
  },

  _moveSiteToEmptyCell: function USTrans__moveSiteToEmptyCell(aCell) {
    this._lock();

    // freeze our site position
    this._freezeSitePosition(this._site);

    // move site the its new DOM position
    aCell.node.appendChild(this._site.node);

    let self = this;

    // slide site into its new position
    this._slideSiteTo(this._site, aCell, function () {
      self._unfreezeSitePosition(self._site);
      self._finish();
    });
  },

  _finish: function USTrans__finish() {
    this._unlock();

    if (this._callback)
      this._callback();
  }
});

// TODO
function DragSiteTransformation(aSite, aEvent, aCallback) {
  this._site = aSite;
  this._callback = aCallback;

  this._start(aEvent);
}

DragSiteTransformation.prototype = extend(BaseTransformation, {
  get _cellPositions() {
    let excludeCell = this._site.cell;

    let cells = Grid.cells.filter(function (cell) {
      return cell != excludeCell;
    });

    let self = this;

    let cellPositions = cells.map(function (cell) {
      return {cell: cell, rect: self._getNodePosition(cell.node)};
    });

    // replace the getter with our cached value
    let getter = function DSTrans__getCellPositions() cellPositions;
    this.__defineGetter__("_cellPositions", getter);
    return cellPositions;
  },

  _start: function DSTrans__start(aEvent) {
    // listen for 'dragend'
    let node = this._site.node;
    this._dragend = this._dragend.bind(this);
    node.addEventListener("dragend", this._dragend, false);

    // listen for 'dragover'
    this._dragover = this._dragover.bind(this);
    document.addEventListener("dragover", this._dragover, false);

    // store cursor offset
    let rect = node.getBoundingClientRect();
    this._offsetX = aEvent.clientX - rect.left;
    this._offsetY = aEvent.clientY - rect.top;

    // store cell dimensions
    let cellNode = this._site.cellNode;
    this._cellWidth = cellNode.offsetWidth;
    this._cellHeight = cellNode.offsetHeight;

    // store minimum intersection values
    this._minWidth = this._cellWidth / 2;
    this._minHeight = this._cellHeight / 2;

    // TODO
    let node = this._site.node;
    node.classList.add("dragged");

    this._freezeSitePosition(this._site);
  },

  _dragover: function DSTrans__dragover(aEvent) {
    // get the viewport size
    let {clientWidth, clientHeight} = document.documentElement;

    // we'll want a padding of 5px
    let border = 5;

    // TODO add some good variable names
    // enforce minimum constraints to keep the drag image inside the window
    let left = Math.max(scrollX + aEvent.clientX - this._offsetX, border);
    let top = Math.max(scrollY + aEvent.clientY - this._offsetY, border);

    // TODO add some good variable names
    // enforce maximum constraints to keep the drag image inside the window
    left = Math.min(left, scrollX + clientWidth - this._cellWidth - border);
    top = Math.min(top, scrollY + clientHeight - this._cellHeight - border);

    // update the drag image's position
    this._setSitePosition(this._site, {left: left, top: top});

    // find a possible drop target for the current position
    let targetCell = this._findDropTarget(aEvent, this._site.cell);

    // TODO
    if (this._lastDropTarget != targetCell) {
      if (this._rearrangeTimeout)
        clearTimeout(this._rearrangeTimeout);

      let self = this;
      let callback = function () self._rearrange(targetCell);
      this._rearrangeTimeout = setTimeout(callback, 250);
    }

    this._lastDropTarget = targetCell;
  },

  _rearrange: function DSTrans__rearrange(aTarget) {
    let targetSite = aTarget && aTarget.site;
    let lastTargetSite = this._lastTarget && this._lastTarget.site;

    if (lastTargetSite)
      // TODO finish must not be called before this has finished
      this._slideSiteTo(lastTargetSite, this._lastTarget);

    if (targetSite) {
      this._freezeSitePosition(targetSite);
      // TODO finish must not be called before this has finished
      this._slideSiteTo(targetSite, this._site.cell);
    }

    this._lastTarget = aTarget;
  },

  _findDropTarget: function DSTrans__findDropTarget(aEvent) {
    let {_minWidth: minWidth, _minHeight: minHeight} = this;

    let rect = new Rect(aEvent.clientX - this._offsetX,
                        aEvent.clientY - this._offsetY,
                        this._cellWidth, this._cellHeight);

    let targetCell;

    this._cellPositions.some(function (pos) {
      let inter = rect.intersect(pos.rect);

      if (inter.width >= minWidth && inter.height >= minHeight) {
        targetCell = pos.cell;
        return true;
      }

      return false;
    });

    return targetCell;
  },


  _dragend: function DSTrans__dragend(aEvent) {
    // remove 'dragend' listener
    let node = this._site.node;
    node.removeEventListener("dragend", this._dragend, false);

    // remove 'dragover' listener
    document.removeEventListener("dragover", this._dragover, false);

    // TODO
    if (this._rearrangeTimeout)
      clearTimeout(this._rearrangeTimeout);

    // TODO
    node.classList.remove("dragged");

    this._lock();

    let finish = this._finish.bind(this);

    if (this._lastDropTarget) {
      let targetSite = this._lastDropTarget.site;

      if (targetSite) {
        // TODO
        this._freezeSitePosition(targetSite);
        this._site.cellNode.appendChild(targetSite.node);
        this._slideSiteTo(targetSite, this._site.cell);

        // TODO
        targetSite.unpin();
        targetSite.pin();
      } else {
        let cell = this._site.cell;

        // if we dragged the site onto an empty cell there might now be a gap
        // in the grid and we have to pin all cells after the gap
        while (cell = cell.nextSibling) {
          let site = cell.site;
          site && site.pin();
        }
      }

      // slide the dragged site into its place and pin it
      this._lastDropTarget.node.appendChild(this._site.node);
      this._slideSiteTo(this._site, this._lastDropTarget, finish);

      // TODO
      this._site.unpin();
      this._site.pin();
    } else {
      // snap back to original place
      this._slideSiteTo(this._site, this._site.cell, finish);
    }
  },

  _finish: function DSTrans__finish() {
    let self = this;

    // unfreeze all sites
    Grid.cells.forEach(function (cell) {
      let site = cell.site;
      if (site)
        self._unfreezeSitePosition(site);
    });

    this._unlock();

    if (this._callback)
      this._callback();
  }
});
//@line 52 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\base\content\newtab\newTab.js"

// TODO
function debug(aMsg) {
  aMsg = ("NewTabPage: " + aMsg).replace(/\S{80}/g, "$&\n");
  Services.console.logStringMessage(aMsg);
}

// TODO
function trace(aMsg) {
  // cut off the first line of the stack trace, because that's just this function.
  let stack = Error().stack.split("\n").slice(1);

  debug("trace: " + aMsg + "\n" + stack.join("\n"));
}

// TODO
function extend(aParent, aObject) {
  for (let name in aParent) {
    let getter, setter;

    if (getter = aParent.__lookupGetter__(name))
      aObject.__defineGetter__(name, getter);
    else if (setter = aParent.__lookupSetter__(name))
      aObject.__defineSetter__(name, setter);
    else
     aObject[name] = aParent[name];
  }

  return aObject;
}

// TODO
let Page = {
  init: function Page_init(aToolbarSelector, aGridSelector) {
    Toolbar.init(aToolbarSelector);
    this._gridSelector = aGridSelector;

    // check if the new tab feature is enabled
    if (this.isEnabled())
      this._init();
    else
      document.body.classList.add("disabled");

    // register pref observer
    Services.prefs.addObserver(PREF_NEWTAB_ENABLED, this, true);
  },

  lock: function Page_lock() {
    document.body.classList.add("locked");
  },

  unlock: function Page_unlock() {
    document.body.classList.remove("locked");
  },

  isLocked: function Page_isLocked() {
    return document.body.classList.contains("locked");
  },

  isEnabled: function Page_isEnabled() {
    return Services.prefs.getBoolPref(PREF_NEWTAB_ENABLED, true);
  },

  modify: function Page_modify() {
    document.body.classList.add("modified");
  },

  reset: function Page_reset() {
    document.body.classList.remove("modified");
  },

  isModified: function Page_isModified() {
    let blockedLinks = NewTabUtils.Storage.get("blockedLinks", {});
    return Object.keys(blockedLinks).length > 0;
  },

  refresh: function Page_refresh() {
    Links.reset();
    Grid.refresh();
  },

  observe: function Page_observe() {
    let classes = document.body.classList;

    if (this.isEnabled()) {
      classes.remove("disabled");
      this._init();
    } else {
      classes.add("disabled");
    }
  },

  _init: function Page__init() {
    if (this._initialized)
      return;

    this._initialized = true;

    NewTabUtils.Pages.register(this);

    let self = this;

    // listen for 'unload' to unregister this page
    addEventListener("unload", function () {
      NewTabUtils.Pages.unregister(self);
    }, false);

    // check if the grid has been modified
    if (this.isModified())
      this.modify();

    Grid.init(this._gridSelector);
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver,
                                         Ci.nsISupportsWeakReference])
};

// TODO
let Toolbar = {
  init: function Toolbar_init(aSelector) {
    this._node = document.querySelector(aSelector);

    let btnShow = this._node.firstElementChild;
    let btnHide = btnShow.nextElementSibling;
    let btnRefresh = btnHide.nextElementSibling;

    btnShow.addEventListener("click", this._show.bind(this), false);
    btnHide.addEventListener("click", this._hide.bind(this), false);
    btnRefresh.addEventListener("click", this._refresh.bind(this), false);
  },

  _show: function Toolbar__show() {
    Services.prefs.setBoolPref(PREF_NEWTAB_ENABLED, true);
  },

  _hide: function Toolbar__hide() {
    Services.prefs.setBoolPref(PREF_NEWTAB_ENABLED, false);
  },

  _refresh: function Toolbar__refresh() {
    Grid.reset();
  }
};

// TODO
let Grid = {
  get cells() {
    let cells = [];
    let children = this._node.querySelectorAll("li");
    let numChildren = children.length;

    // create the cells
    for (let i = 0; i < numChildren; i++)
      cells.push(new Cell(children[i]));

    // replace the getter with our cached value
    this.__defineGetter__("cells", function Grid_getCells() cells);
    return cells;
  },

  get _siteFragment() {
    let site = document.createElement("a");
    site.classList.add("site");
    site.setAttribute("draggable", "true");

    site.innerHTML =
      '<img class="img" width="201" height="127" ' +
      ' src="chrome://browser/content/newtab/thumb.png" alt=""/>' +
      '<span class="title"/>' +
      '<span class="strip">' +
      '  <input class="btn-pin" type="button"/>' +
      '  <input class="btn-block" type="button"/>' +
      '</span>';

    let fragment = document.createDocumentFragment();
    fragment.appendChild(site);

    // replace the getter with our cached value
    let getter = function Grid__getSiteFragment() fragment;
    this.__defineGetter__("_siteFragment", getter);
    return fragment;
  },

  init: function Grid_init(aSelector) {
    this._node = document.querySelector(aSelector);
    this._draw();
  },

  createSite: function Grid_createSite(aLink, aCell) {
    let node = aCell.node;
    node.appendChild(this._siteFragment.cloneNode(true));
    return new Site(node.firstElementChild, aLink);
  },

  reset: function Grid_reset() {
    let node = this._node;

    Animations.fadeOut(node, function () {
      NewTabUtils.Storage.clear();
      NewTabUtils.Pages.reset();
      NewTabUtils.Pages.refresh(Page);
      Page.refresh();

      Animations.fadeIn(node);
    });
  },

  refresh: function Grid_refresh() {
    // remove all sites
    this.cells.forEach(function (cell) {
      let node = cell.node;
      let child = node.firstElementChild;

      if (child)
        node.removeChild(child);
    });

    // draw the grid again
    this._draw();
  },

  _draw: function Grid__draw() {
    let self = this;
    let cells = this.cells;

    // put sites into the cells
    Links.getLinks(function (links) {
      let maxIndex = Math.min(links.length, cells.length);

      for (let i = 0; i < maxIndex; i++)
        if (links[i])
          self.createSite(links[i], cells[i]);
    });
  }
};

// TODO
function Cell(aNode) {
  this._node = aNode;
  this._node.__newtabCell = this;
}

Cell.prototype = {
  get node() this._node,
  get index() Grid.cells.indexOf(this),

  get nextSibling() {
    let next = this.node.nextElementSibling;
    return next && next.__newtabCell;
  },

  get site() {
    let firstChild = this.node.firstElementChild;
    return firstChild && firstChild.__newtabSite;
  },

  containsPinnedSite: function Cell_containsPinnedSite() {
    let site = this.site;
    return site && site.isPinned();
  },

  isEmpty: function Cell_isEmpty() {
    return !this.site;
  }
};

// TODO
function Site(aNode, aLink) {
  this._node = aNode;
  this._node.__newtabSite = this;

  this._link = aLink;

  this._fillNode();
  this._connectButtons();

  this._dragHandler = new DragHandler(this);
}

Site.prototype = {
  get node() this._node,
  get index() this.cell.index,
  get url() this._link.url,
  get title() this._link.title,
  get cellNode() this.cell.node,
  get cell() this.node.parentNode.__newtabCell,

  pin: function Site_pin() {
    if (this.isPinned())
      return;

    this.node.classList.add("pinned");
    this._link.pin(this.index);
  },

  unpin: function Site_unpin() {
    if (!this.isPinned())
      return;

    this.node.classList.remove("pinned");
    this._link.unpin();
  },

  isPinned: function Site_isPinned() {
    return this._link.isPinned();
  },

  block: function Site_block(aEvent) {
    this._link.block();
  },

  _querySelector: function (aSelector) {
    return this.node.querySelector(aSelector);
  },

  _fillNode: function Site__fillNode() {
    let {url, title} = this;

    this.node.setAttribute("href", url);
    this._querySelector(".img").setAttribute("alt", title);
    this._querySelector(".title").textContent = title;

    if (this.isPinned())
      this.node.classList.add("pinned");
  },

  _connectButtons: function Site__connectButtons() {
    let self = this;

    this._querySelector(".btn-pin").addEventListener("click", function (aEvent) {
      if (aEvent)
        aEvent.preventDefault();

      if (self.isPinned()) {
        self.unpin();
        Transformation.unpinSite(self);
      } else {
        self.pin();
      }
    }, false);

    this._querySelector(".btn-block").addEventListener("click", function (aEvent) {
      if (aEvent)
        aEvent.preventDefault();

      if (!Page.isModified())
        NewTabUtils.Pages.modify();

      self.block();
      Transformation.blockSite(self);
    }, false);
  }
};

// TODO
function Link(aUrl, aTitle) {
  this._url = aUrl;
  this._title = aTitle;
}

Link.unserialize = function (aData) {
  return new Link(aData.url, aData.title);
};

Link.prototype = {
  get url() this._url,
  get title() this._title || this._url,

  block: function Link_block() {
    Links.block(this);
  },

  pin: function Link_pin(aIndex) {
    Links.pin(this, aIndex);
  },

  unpin: function Link_unpin() {
    Links.unpin(this);
  },

  isPinned: function Link_isPinned() {
    return Links.isPinned(this);
  },

  serialize: function Link_serialize() {
    return {url: this.url, title: this.title};
  }
};

// TODO
let Links = {
  getLinks: function Links_getLinks(aCallback) {
    let pinnedLinks = this._pinnedLinks.concat();
    let blockedLinks = this._blockedLinks;

    NewTabUtils.Links.fetchLinks(function (links) {
      // create Link objects
      links = links.map(function (link) {
        return new Link(link.url, link.title);
      });

      // filter blocked and pinned links
      links = links.filter(function (link) {
        return !(link.url in blockedLinks) && !link.isPinned();
      });

      // unserialize all pinned links
      for (let i = 0; i < pinnedLinks.length; i++)
        if (pinnedLinks[i])
          pinnedLinks[i] = Link.unserialize(pinnedLinks[i]);

      // try to fill the gaps between pinned links
      for (let i = 0; i < pinnedLinks.length && links.length; i++)
        if (!pinnedLinks[i])
          pinnedLinks[i] = links.shift();

      // append the remaining links if any
      if (links.length)
        pinnedLinks = pinnedLinks.concat(links);

      aCallback(pinnedLinks);
    });
  },

  block: function Links_block(aLink) {
    this._blockedLinks[aLink.url] = 1;
    NewTabUtils.Storage.set("blockedLinks", this._blockedLinks);

    // make sure we unpin blocked links
    if (!this.unpin(aLink))
      NewTabUtils.Pages.refresh(Page);
  },

  pin: function Links_pin(aLink, aIndex) {
    this._pinnedLinks[aIndex] = aLink.serialize();
    NewTabUtils.Storage.set("pinnedLinks", this._pinnedLinks);
    NewTabUtils.Pages.refresh(Page);
  },

  unpin: function Links_unpin(aLink) {
    let index = this._pinnedIndexOf(aLink);

    if (index > -1) {
      this._pinnedLinks[index] = null;
      NewTabUtils.Storage.set("pinnedLinks", this._pinnedLinks);
      NewTabUtils.Pages.refresh(Page);

      return true;
    }

    return false;
  },

  isPinned: function Links_isPinned(aLink) {
    return this._pinnedIndexOf(aLink) > -1;
  },

  reset: function Links_reset() {
    this._initStorageValues();
  },

  _init: function Links__init() {
    this._initStorageValues();
  },

  _initStorageValues: function Links__initStorageValues() {
    this._pinnedLinks = NewTabUtils.Storage.get("pinnedLinks", []);
    this._blockedLinks = NewTabUtils.Storage.get("blockedLinks", {});
  },

  _pinnedIndexOf: function Links__pinnedIndexOf(aLink) {
    let pinned = this._pinnedLinks;
    let url = aLink.url;

    for (let i = 0; i < pinned.length; i++)
      if (pinned[i] && pinned[i].url == url)
        return i;

    return -1;
  }
};

Links._init();

// TODO
function DragHandler(aSite) {
  this._site = aSite;
  this._node = this._site.node;

  // listen for 'dragstart'
  this._dragstart = this._dragstart.bind(this);
  this._node.addEventListener("dragstart", this._dragstart.bind(this), false);
}

DragHandler.prototype = {
  _dragstart: function DragHandler__dragstart(aEvent) {
    let node = this._node;

    // TODO
    node.classList.add("ontop");
    document.body.classList.add("dragging");

    let self = this;

    Transformation.dragSite(this._site, aEvent, function () {
      node.classList.remove("ontop");
      document.body.classList.remove("dragging");
      document.body.removeChild(self._dragImage);
    });

    this._setDragData(aEvent);

    return true;
  },

  _setDragData: function DragHandler__setDragData(aEvent) {
    let {url, title} = this._site;

    let dt = aEvent.dataTransfer;
    dt.setData("text/plain", url);
    dt.setData("text/uri-list", url);
    dt.setData("text/x-moz-url", url + "\n" + title);
    dt.setData("text/html", "<a href=\"" + url + "\">" + url + "</a>");

    // create and use an empty drag image
    let img = this._dragImage = document.createElement("div");
    img.classList.add("drag-img");
    document.body.appendChild(img);
    dt.setDragImage(img, 0, 0);

    dt.effectAllowed = "move";
  }
};
